"""
plot_compare.py

This script plots 1+ timeseries figures of simulation data with overlain reference data.

Reference data without provincial stratification should have a 'Province' column with the value 'All'
"""

import itertools
import math
import matplotlib.pyplot as plt
import numpy as np
import operator
import os
import pandas as pd
import sys

from calibtool.CalibManager import CalibManager
from dtk.utils.observations.AgeBin import AgeBin
from dtk.utils.observations.BaseDistribution import BaseDistribution
from dtk.utils.observations.DataFrameWrapper import DataFrameWrapper
from dtk.utils.observations.PopulationObs import PopulationObs
from dtk.utils.observations.utils import parse_ingest_data_from_xlsm
from matplotlib import collections as mc
from simtools.Analysis.AnalyzeManager import AnalyzeManager
from simtools.Analysis.BaseAnalyzers.DownloadAnalyzer import DownloadAnalyzer
from simtools.SetupParser import SetupParser
from simtools.Utilities.COMPSUtilities import exps_for_suite_id
from simtools.Utilities.Experiments import retrieve_experiment
import matplotlib.patches as mpatches
from hiv.lib.utils.analysis import model_population_in_year

FIG_HEIGHT = 20
FIG_WIDTH = 10
TICK_SPACING = 3

MOST_LIKELY_MODE = 'most-likely'
LEAST_LIKELY_MODE = 'least-likely'
RANDOM_MODE = 'random'
SELECTION_MODES = [MOST_LIKELY_MODE, LEAST_LIKELY_MODE, RANDOM_MODE]


class Colors:
    male = 'blue'
    female = 'red'
    both = 'black'


class MissingChannelException(Exception):
    pass


class DownloadIfNotPresentAnalyzer(DownloadAnalyzer):
    """
    Analyzer only downloading the files if not already present.
    Also limits the number of sims retrieved
    """
    def __init__(self, filenames, output_path, **kwargs):
        super().__init__(filenames, output_path, **kwargs)

    def filter(self, simulation):
        return not all(
            os.path.exists(os.path.join(self.output_path, simulation.id, os.path.basename(filename)))
            for filename in self.filenames)


def get_gendered_color(gender):
    try:
        color = getattr(Colors, gender.lower())
    except AttributeError:
        raise Exception('Unknown gender for plot color selection: %s' % gender)
    return color


def get_candidate_simulations(args):
    if args.object_id is not None:
        # Retrieve the object (either Experiment or suite)
        try:
            exp_list = [retrieve_experiment(args.object_id)]
        except:
            try:
                exp_list = exps_for_suite_id(args.object_id)
                exp_list = [retrieve_experiment(e.id) for e in exp_list]
            except:
                print('The provided id: %s is neither an experiment nor a suite... Exiting.' % args.object_id)
                exit()
        simulation_ids = []
        for exp in exp_list:
            simulation_ids.append([s.id for s in exp.simulations])
        candidates = list(itertools.chain(*simulation_ids))
    elif args.calibration_directory:
        cm = CalibManager.open_for_reading(calibration_directory=args.calibration_directory)
        candidates = cm.get_parameter_sets_with_likelihoods()  # returns ParameterSet objects
    else:
        raise Exception('Should not be here, talk to a dev as this is a bug.')
    return candidates


def select_simulation_ids_to_plot(candidates, criterion, n_selections, output_directory):
    if criterion == RANDOM_MODE:
        # candidates can either be sim ids or ParameterSet objects
        selected_samples = list(np.random.choice(candidates, size=n_selections, replace=False))
        try:
            selected_simulation_ids = [ps.sim_id for ps in selected_samples]
            can_write_samples = True
        except AttributeError:
            selected_simulation_ids = selected_samples
            can_write_samples = False  # all we have are simulation ids with no parameter value-related information
    else:
        # candidates must be ParameterSet objects with likelihood set
        reverse = True if criterion == MOST_LIKELY_MODE else False
        sorted_candidates = sorted(candidates, key=lambda ps: ps.likelihood_exponentiated, reverse=reverse)
        selected_samples = sorted_candidates[0:n_selections]
        selected_simulation_ids = [ps.sim_id for ps in selected_samples]
        can_write_samples = True

    # write out the selected samples in case someone wants to see/use them later
    if can_write_samples:
        samples_df = pd.DataFrame(data=[s.to_dict() for s in selected_samples])
        samples_df.to_csv('plot_compare_parameter_sets.csv')
    else:
        print('Could not write plot_compare_parameter_sets.csv as no parameter value information is available.')
    return selected_simulation_ids


def download_files(simulation_ids, filenames, output_directory):
    analyzer = DownloadIfNotPresentAnalyzer(filenames=filenames, output_path=output_directory)
    am = AnalyzeManager(sim_list=simulation_ids, analyzers=analyzer)
    am.analyze()
    return am.simulations


def read_sim_data(fn, node_map):
    sim_df = pd.read_csv(fn).sort_values(by='Year', ascending=True)

    # fix up Node-Province column
    if 'Node' in list(sim_df.columns):
        sim_df = sim_df.set_index('Node').rename(node_map).reset_index().rename(columns={'Node': 'Province'})
    return PopulationObs(dataframe=sim_df)


def make_collection(d, x='Year', y='Result'):
    try:
        zipped_data = zip(d[x], d[y])
    except TypeError:
        zipped_data = zip([d[x]], [d[y]])
    return zipped_data


def scale_to_census(df, pop_df, census_population, census_year, census_min_age, census_max_age, verbose=False):
    model_pop, census_to_model_ratio = model_population_in_year(year=census_year, obs_population=census_population,
                                                                df=pop_df,
                                                                low_age=census_min_age, high_age=census_max_age,
                                                                population_col='Result', verbose=verbose)
    df = df.assign(Result=df['Result'] * census_to_model_ratio)
    return df


def make_and_plot_collection(groups, provinces, figure_dict, channel, **kwargs):
    for group_tuple, data in groups:
        age_bin, province, gender = group_tuple
        if province in [1, '1']:  # kludgy fix for single-node simulations
            continue
        color = get_gendered_color(gender=gender)
        alpha = kwargs.get('alpha', 1.0)
        marker = kwargs.get('marker', '.')
        linewidth = kwargs.get('linewidth', 0.5)

        # Try to obtain the axis to add data to and create it on the fly if it does not exist
        if age_bin in figure_dict:
            fd = figure_dict[age_bin]
        else:
            figure, axis = plt.subplots(max(len(provinces), 2), 1, sharex=True, sharey=True,
                                        figsize=(FIG_WIDTH, FIG_HEIGHT))
            figure_dict[age_bin] = {'figure': figure, 'axis': axis}
            fd = figure_dict[age_bin]

        axis = fd['axis']
        province_index = provinces.index(province)

        # get simulation data and plot it
        if len(data.index) == 1:
            # a single data element
            axis[province_index].plot(data['Year'], data[channel], marker=marker, color=color, alpha=alpha, ms=15)
        else:
            # a line of data
            coll = list(make_collection(data, x='Year', y=channel))
            lc = mc.LineCollection([coll], linewidths=linewidth)
            lc.set_alpha(alpha)
            lc.set_color(color)
            axis[province_index].add_collection(lc)


def make_and_plot_bxp(groups, genders, provinces, figure_dict, channel, lower_channel, upper_channel):
    """
    Make a center-point and whisker plot (subset of a boxplot) for reference data.
    Args:
        groups: dataframe.groupby() result
        genders: list of genders available (needed for subplot identification)
        provinces: list of provinces available (needed for subplot identification)
        figure_dict: dict of figures and axes for plotting on
        channel: data channel name to plot
        lower_channel: data channel name of the lower-bound uncertainty of the central channel
        upper_channel: data channel name of the upper-bound uncertainty of the central channel

    Returns: no return
    """
    for group_tuple, data in groups:
        age_bin, province, gender = group_tuple

        fd = figure_dict[age_bin]
        axis = fd['axis']

        province_index = provinces.index(province)
        bxp_tuples = zip(data[lower_channel], data[channel], data[upper_channel])
        bxp_dicts = [
            {
                'q1': tup[0],  # required but dummy, will turn off
                'q3': tup[0],  # required but dummy, will turn off
                'whislo': tup[0],
                'med': tup[1],
                'whishi': tup[2],
            }
            for tup in bxp_tuples
        ]

        # Depending on gender, draw differently
        gender_index = genders.index(gender)
        gender_offset = TICK_SPACING / 5 if gender_index == 1 else -TICK_SPACING / 5
        positions = [p + gender_offset for p in data['Year']]
        color = get_gendered_color(gender=gender)
        drawing_props = dict(color=color)

        # Draw on the axis
        axis[province_index].bxp(bxpstats=bxp_dicts, showbox=False, showfliers=False, showmeans=False,
                                 positions=positions, medianprops=drawing_props, whiskerprops=drawing_props,
                                 capprops=drawing_props)


def generate_plot(reference, sim_filenames, pop_filenames, node_map, channel, census_population, census_year,
                  census_min_age, census_max_age, scaling, distribution, start_year=None, end_year=None, verbose=False):
    # two-tailed two sigma error bars, p = 0.02275 and 0.97725 applied regardless of distribution
    p_low = (1 - 0.9545) / 2
    p_high = 1 - p_low

    p_low_channel = reference.add_percentile_values(channel=channel, distribution=distribution, p=p_low)[0]
    p_high_channel = reference.add_percentile_values(channel=channel, distribution=distribution, p=p_high)[0]

    # data always has provinces now; even if just the 'All' province
    reference_provinces = reference.get_provinces()
    for province in reference_provinces:
        if isinstance(province, str):
            continue
        if math.isnan(province):
            raise DataFrameWrapper.MissingRequiredData('')

    allowed_genders = ['Male', 'Female', 'Both']
    genders = [gender for gender in reference.get_genders() if gender in allowed_genders]

    # the rest of this method expects reference to be a multi-indexed DataFrame
    reference = reference._dataframe

    # population scaling for sim data, T/F
    if scaling:
        age_bin = AgeBin(start=census_min_age, end=census_max_age)
        print('-- Will perform scaling of channel: %s sim data to census population, age bin: %s.' % (channel, age_bin))
    else:
        print('-- No population scaling will be performed for channel: %s' % channel)

    # determine time bounds for plotted data
    time_conditions = []
    if start_year:
        time_conditions.append(['Year', operator.ge, start_year])
    if end_year:
        time_conditions.append(['Year', operator.le, end_year])

    # used for to group data for plotting
    multi_index = ['AgeBin', 'Province', 'Gender']

    n_sims = len(sim_filenames)
    n = 1
    figure_dict = {}
    all_years = set()  # we will use this for setting year ticks
    for i in range(len(sim_filenames)):
        fn = sim_filenames[i]

        if verbose:
            print('Processing sim file %d/%d ...' % (n, n_sims))
            sys.stdout.flush()
        results = read_sim_data(fn, node_map=node_map).filter(conditions=time_conditions)
        all_years = all_years.union(set(results.get_years()))
        results = results._dataframe

        # scale Results to census data
        if scaling:
            pop_filename = pop_filenames[i]
            pop = read_sim_data(pop_filename, node_map=node_map)._dataframe
            results = scale_to_census(results, pop, census_population, census_year, census_min_age, census_max_age,
                                      verbose=verbose)

        # determine the union of sim and reference provinces (their intersection can be empty, all, or between)
        sim_provinces = results['Province'].unique()  # ck4, define
        provinces = list(set(reference_provinces).union(set(sim_provinces)))
        # putting aggregated node at the end (RHS) of all plots
        if PopulationObs.AGGREGATED_PROVINCE in provinces:
            provinces.remove(PopulationObs.AGGREGATED_PROVINCE)
        provinces.append(PopulationObs.AGGREGATED_PROVINCE)

        make_and_plot_collection(groups=results.groupby(multi_index), provinces=provinces,
                                 channel='Result', figure_dict=figure_dict,
                                 alpha=0.1, linewidth=0.5, marker='.')
        n += 1

    # independent/year axis ticks
    ticks_start = math.ceil(min(all_years) / TICK_SPACING) * TICK_SPACING
    ticks_end = math.ceil(
        (max(all_years) + 1e-3) / TICK_SPACING) * TICK_SPACING  # one too far, but the range fixes this
    ticks = list(range(ticks_start, ticks_end, TICK_SPACING))

    # reference data point-and-whisker plots
    print('Plotting reference data...')
    groups = reference.groupby(multi_index)
    make_and_plot_bxp(groups=groups, genders=genders, provinces=provinces,
                      channel=channel, lower_channel=p_low_channel, upper_channel=p_high_channel,
                      figure_dict=figure_dict)

    # finalize figures
    print('Finalizing figures...')
    sys.stdout.flush()
    for age_bin, fd in figure_dict.items():
        # set axis ticks & labels and make the plot pretty
        axis = fd['axis']
        for province_index in range(len(provinces)):
            axis[province_index].set_ylim(bottom=0)  # bottom of the plot should be 0
            axis[province_index].set_xticks(ticks)
            axis[province_index].set_xticklabels(ticks, rotation=75)

            if province_index == 0:
                axis[province_index].set_title('%s %s' % (channel, age_bin), fontsize=12)

            axis[province_index].set_ylabel('%s' % provinces[province_index].replace(' ', '\n'), fontsize=8)

            if province_index == len(provinces) - 1:
                axis[province_index].set_xlabel('Year')
                female_patch = mpatches.Patch(color=Colors.female, label='Female')
                male_patch = mpatches.Patch(color=Colors.male, label='Male')
                both_patch = mpatches.Patch(color=Colors.both, label='Both')
                axis[province_index].legend(handles=[female_patch, male_patch, both_patch], loc='upper right')

        figure = fd['figure']
        figure.savefig('%s/%s_%s.png' % (args.output_directory, channel, age_bin.replace(':', '-')), bbox_inches='tight')


def detect_uncertainty_channel(dfw, channel):
    # the uncertainty channel should be the only channel that is not a stratifier or 'weight' or the channel itself
    df = dfw._dataframe.dropna(subset=[channel.name])
    df = df.dropna(axis='columns')
    possible_channels = list(set(df.columns) - set(dfw.stratifiers) - set([channel.name, dfw.WEIGHT_CHANNEL]))
    if len(possible_channels) != 1:
        raise Exception('Cannot determine the uncertainty channel. Possibilities: %s' % possible_channels)
    return possible_channels[0]


def main(args):
    # get reference data and site info from ingest form
    params, site_info, reference, analyzers, channels = parse_ingest_data_from_xlsm(filename=args.reference_file)

    try:
        channel = next(ch for ch in channels if ch.name == args.channel)
        analyzer = next(a for a in analyzers if a['channel'] == args.channel)
    except Exception:
        raise MissingChannelException('No such channel: %s in ingest form: %s' % (args.channel, args.reference_file))

    # obtain output data file to get plotting data from
    comps_output_root = os.path.join('output', 'post_process')

    sim_filename = '%s.csv' % args.channel
    comps_filenames = [os.path.join(comps_output_root, sim_filename)]

    uncertainty_channel = detect_uncertainty_channel(dfw=reference, channel=channel)
    reference = reference.filter(keep_only=[args.channel, uncertainty_channel])

    distribution = BaseDistribution.from_uncertainty_channel(uncertainty_channel=uncertainty_channel)
    distribution.prepare(dfw=reference, channel=args.channel)

    # add in Population for computing the pop scaling factor
    if channel.needs_pop_scaling:
        print('Pop scaling will occur')
        comps_filenames.append(os.path.join(comps_output_root, 'Population.csv'))
    else:
        print('Pop scaling will NOT occur')

    # determine which simulations will be plotted
    candidates = get_candidate_simulations(args)
    simulation_ids = select_simulation_ids_to_plot(candidates=candidates, criterion=args.selection_mode,
                                                   n_selections=args.n_sims, output_directory=args.output_directory)
    simulation_ids = download_files(simulation_ids=simulation_ids, filenames=comps_filenames,
                                    output_directory=args.output_directory)

    # create requested plots
    # Make sure these two lists are in the same sim order
    result_filenames = [os.path.join(args.output_directory, sim_id, sim_filename) for sim_id in simulation_ids]
    population_filenames = [os.path.join(args.output_directory, sim_id, 'Population.csv') for sim_id in simulation_ids]

    min_age = site_info['census_age_bin'].start
    max_age = site_info['census_age_bin'].end
    node_map = site_info['node_map']
    # add the non-aggregated node to the node map
    node_map[PopulationObs.AGGREGATED_NODE] = PopulationObs.AGGREGATED_PROVINCE

    generate_plot(reference, result_filenames, population_filenames, node_map=node_map, channel=args.channel,
                  census_population=site_info['census_population'], census_year=site_info['census_year'],
                  census_min_age=min_age, census_max_age=max_age, scaling=channel.needs_pop_scaling,
                  distribution=distribution, start_year=args.start_year, end_year=args.end_year, verbose=args.verbose)
    plt.show()


def parse_args():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--channel', dest='channel', type=str, required=True,
                        help='Channel to plot (Required).')
    parser.add_argument('-r', '--reference', dest='reference_file', type=str, required=True,
                        help='Ingest form xlsm file to use (Required).')
    parser.add_argument('-id', dest='object_id', type=str, default=None,
                        help='Experiment/Suite id from which to gather simulation data for plotting'
                             '(Mutually exclusive with -d).')
    parser.add_argument('-d', dest='calibration_directory', type=str, default=None,
                        help='Experiment/Suite id from which to gather simulation data for plotting'
                             '(Mutually exclusive with -id).')
    parser.add_argument('-o', '--output_directory', dest='output_directory', default='output',
                        help='Directory to read/write simulation data from (Default: output).')
    parser.add_argument('-n', '--nsims', dest='n_sims', type=int, default=100,
                        help='Plot this number of simulations in the experiment (Default: (up to the) first 100).')
    parser.add_argument('-m', '--mode', dest='selection_mode', type=str, required=True, choices=SELECTION_MODES,
                        help='Criterion for selecting simulations to plot (Required).')

    parser.add_argument('--start_year', dest='start_year', type=float, default=None,
                        help='Plot data starting at this inclusive year (Default: beginning of all data).')
    parser.add_argument('--end_year', dest='end_year', type=float, default=None,
                        help='Plot data through this inclusive year (Default: end of all data).')
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        help='Print more info to the screen during processing.')

    args = parser.parse_args()
    return args


def validate_args(args):
    # Must specify one of these two arguments; not both, not zero
    if not((args.object_id is None) ^ (args.calibration_directory is None)):
        raise Exception('Must specify either a calibration directory (-d) or suite/experiment id (-id)')

    # there is no likelihood information available in -id mode so it can only use the random selection mode
    if args.object_id is not None:
        if args.selection_mode != RANDOM_MODE:
            raise Exception('Only selection mode: %s is available for use when supplying an -id for plotting.' %
                            RANDOM_MODE)


if __name__ == '__main__':
    args = parse_args()
    validate_args(args)

    SetupParser.init(selected_block='HPC')

    plt.tight_layout()
    main(args)
